from __future__ import absolute_import
from future import standard_library
standard_library.install_aliases()
from builtins import zip
from builtins import object
from .cfg_lib import parse_config_iter
import logging
from itertools import cycle, zip_longest

logger = logging.getLogger(__name__)

default_converters = {'float': float, 'int': int}



"""
converters (int and float) by adding :
defaults (only at end of line) by =
varargs using *, only at end
cannot have both varargs and defaults
no spaces around *, : and =
Line is of form:
KEYWORD1 par1 par2:float par3=33:int par4=bla
KEYWORD2 par1 *weights:float *inputs
"""

class CfgLine(object):
    def __init__(self, key, fname=None, iline=None):
        self.key = key
        self._iline = iline
        self._fname = fname

    def __repr__(self):
        s = """{} defined on line {} of {}""".format(
            self.key, self._iline, self._fname
        )
        s += ''.join('\n  {}: {}'.format(k, v)
                     for k, v in self.__dict__.items()
                     if k != 'key' and not k.startswith('_'))
        return s


def _parse_syntax(template, converters):
    template = template.split()
    key = template[0]
    normal_params = []
    default_args = []
    var_params = []

    found_default = False
    found_vararg = False

    for param in template[1:]:
        param, colon, conv = param.partition(':')

        # strip off converter function
        if colon:
            conv = converters[conv]  # KeyError for unknown converter
        else:
            conv = None

        assert not ('*' in param and '=' in param), \
            'attribute cannot have both default and be a vararg'

        if param.startswith('*'):
            found_vararg = True
            param = param[1:]
            var_params.append((param, conv))
        else:
            assert not found_vararg, 'normal parameter after variable one'

            # strip off default argument
            param, eq, default = param.partition('=')
            if eq:
                found_default = True
            else:
                assert not found_default, \
                    'parameter without default after one with default'
                default = None

            normal_params.append((param, conv))
            default_args.append(default)

    return key, normal_params, default_args, var_params

_template_cache = {}

def parse_line(template, converters, split_line, fname=None, iline=None):
    if template not in _template_cache:
        _template_cache[template] = _parse_syntax(template, converters)
    key, normal_params, default_args, var_params = _template_cache[template]

    if split_line[0] != key:
        raise ValueError('not the right key')

    key, args = split_line[0], split_line[1:]

    line = CfgLine(key, fname, iline)

    n = len(normal_params)
    normal_args, var_args = args[:n], args[n:]

    for (param, conv), value, default in zip_longest(
            normal_params, normal_args, default_args, fillvalue=None):

        if value is None:
            value = default

        if value is None:
            logger.error('Not enough parameters for %s on line %i of %s',
                         key, iline, fname)
            break

        if conv:
            value = conv(value)
        setattr(line, param, value)

    for (param, _) in var_params:
        setattr(line, param, [])

    if var_params and len(var_args) % len(var_params):
        logger.error('number of variable params for %s on line %i of %s'
                  ' must be multiple of %i', key, iline, fname,
                     len(var_params))
        # chop off excess arguments
        n = (len(var_args) // len(var_params)) * len(var_params)
        var_args = var_args[:n]

    for (param, conv), value in zip(cycle(var_params), var_args):
        if conv:
            value = conv(value)
        getattr(line, param).append(value)
    return line


class CfgParser(object):
    def __init__(self, table, skipped_keys = None, converters=default_converters):
        self.templates = {}
        for template in table.splitlines():
            template = template.partition('#')[0]  # remove comments
            if not template:
                continue  # skip blank lines

            key = template.split()[0]
            self.templates[key] = template

        # make set for logging unknown keys only once
        if skipped_keys is None:
            self.skipped_keys = set()
        else:
            self.skipped_keys = skipped_keys
        self.converters = converters

    def parse_file(self, filename):
        lines = []
        for fname, iline, split_line in parse_config_iter(filename):
            key = split_line[0]
            if key not in self.templates:
                if key not in self.skipped_keys:
                    logger.debug('unknown key %s on line %i of %s',
                                 key, iline, fname)
                    self.skipped_keys.add(key)
                continue
            try:
                lines.append(
                    parse_line(self.templates[key], self.converters,
                               split_line, fname, iline))
            except ValueError as err:
                logger.error('Parsing error on line %i of %s: %s',
                             iline, fname, err)
        return lines
